import numpy as np, functools
from functools import reduce
from fastNLP.embeddings.torch.utils import get_embeddings, get_sinusoid_encoding_table
from typing import List, Dict

import inspect
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils import rnn

def make_optim_helper(module, **lr_settings):
    # input: backbone.alpha
    all_param = {k: p for k,p in module.named_parameters() if p.requires_grad==True}
    lr_paramd = []
    for module_name, lr in lr_settings.items():
        params = [p for k,p in all_param.items() if k.startswith(module_name)]
        all_param = {k: v for k,v in all_param.items() if not k.startswith(module_name)}
        lr_paramd.append({'params': params, 'lr': lr, 'name': module_name})
    lr_paramd.append({'params': [p for k, p in all_param.items()], 'name': 'rest'})
    return lr_paramd

def make_optim(model, lr, optim_name='AdamW', add_lr={}, weight_decay=1e-2):
    if optim_name == 'Adam':
        optim = torch.optim.Adam
    elif optim_name == 'SGD':
        optim = functools.partial(torch.optim.SGD, momentum=0.9, weight_decay=0)
    elif optim_name == 'AdamW':
        optim = torch.optim.AdamW
    else:
        raise ValueError('optim not defined')
    # scheduler = torch.optim.lr_scheduler.StepLR(optim, step_size=30, gamma=0.1)
    return optim(make_optim_helper(model, **add_lr), lr=lr)

def get_pos_embed(embed_type, max_position, embed_dim):
    if embed_type == 'sin':
        return nn.Embedding.from_pretrained(
            get_sinusoid_encoding_table(max_position + 1, embed_dim, padding_idx=0),
            freeze=True) 
    elif embed_type == 'learned':
        return get_embeddings((max_position + 1, embed_dim), padding_idx=0)
    else:
        return None


def tween(lst, item, add_last=False):
    """
    >>> a, b = [1,2,3], ['#','$']
    >>> tween(a,b)
    [1, '#', '$', 2, '#', '$', 3]
    >>> tween(a,b,True)
    [1, '#', '$', 2, '#', '$', 3, '#', '$']
    """
    if not isinstance(item, list):
        item = [item]
    if add_last:
        return reduce(lambda r,v: r+[v]+item, lst, [])
    else:
        return reduce(lambda r,v: r+item+[v], lst[1:], lst[:1])

def build_args(func, **kwargs):
    spect = inspect.getfullargspec(func)
    if spect.varkw is not None: 
        return kwargs
    needed_args = set(spect.args)
    if spect.defaults is not None:
        defaults = [arg for arg in spect.defaults]
    else:
        defaults = []
    start_idx = len(spect.args) - len(defaults)
    output = {name: default for name, default in zip(spect.args[start_idx:], defaults)}
    output.update({name: val for name, val in kwargs.items() if name in needed_args})
    return output

def match_args(func, *args, **kwargs) -> dict:
    spect = inspect.signature(func)
    parameters = spect.parameters
    parameters = {
        k:v for k,v in spect.parameters.items() if k not in ('self', 'kwargs')
    }
    spect._parameters = parameters
    args = (*args, *kwargs.pop('args', ()))
    kwargs = {k: v for k,v in kwargs.items() if k in parameters}
    args = spect.bind_partial(*args, **kwargs)
    args.apply_defaults()
    return dict(args.arguments)

class xpack:

    def __init__(self, data, length, batch_first=True, enforce_sorted=False):
        if length.is_cuda: length = length.cpu()
        self.length = length
        self.data, self.batch_sizes, self.sorted_indices, self.unsorted_indices = rnn.pack_padded_sequence(data, length, batch_first=batch_first, enforce_sorted=enforce_sorted)
        self.extra = {}   # to help memorizing other stuff
    
    @property
    def sequence(self):
        return rnn.PackedSequence(
            data=self.data, batch_sizes=self.batch_sizes,
            sorted_indices=self.sorted_indices, unsorted_indices=self.unsorted_indices
        )
    
    def pad(self, data=None, batch_first=True, padding_value=0, total_length=None, **kwargs):
        if data is None: data = self.data
        if not isinstance(data, rnn.PackedSequence):
            batch_sizes = kwargs.get('batch_sizes', self.batch_sizes)
            sorted_indices = kwargs.get('sorted_indices', self.sorted_indices)
            unsorted_indices = kwargs.get('unsorted_indices', self.unsorted_indices)
            data = rnn.PackedSequence(
                data, batch_sizes, sorted_indices, unsorted_indices
            )
        rtn, _ = rnn.pad_packed_sequence(
            data, batch_first=batch_first, padding_value=padding_value, total_length=total_length
        )
        return rtn

    def pack(self, data, batch_first=True, enforce_sorted=False, return_sequence=False):
        seq =  rnn.pack_padded_sequence(data, self.length, batch_first=batch_first, enforce_sorted=enforce_sorted)
        return seq if return_sequence else seq.data

    def repeat(self, data):
        idx = torch.cat([self.sorted_indices[:i] for i in self.batch_sizes])
        if not isinstance(data, torch.Tensor):
            idx = idx.cpu().numpy()
        return data[idx]
    
    def np_pack(self, data):
        idx, size = self.sorted_indices.cpu().numpy(), self.batch_sizes.numpy()
        rtn = []
        for i, size in enumerate(self.batch_sizes):
            rtn.extend(data[idx[:size],i])
        return np.array(rtn)

class Model(nn.Module):

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        instance._init_params_d = match_args(instance.__init__, *args, **kwargs)
        return instance

    @property
    def init_parms_d(self):
        return self._init_params_d

    @classmethod
    def restore_from_checkpoint(cls, checkpoint):
        try:
            ckpt = torch.load(checkpoint)
            model = cls(**ckpt['init_params'])
            model.load_state_dict(ckpt['state_dict'], strict=False)
            return model
        except OSError: 
            pass # start training from scratch
    
    def save_to_checkpoint(self, ckpt_fp, **metadata):
        torch.save({
            'metadata': metadata,
            'state_dict': self.state_dict(),
            'init_params': self.init_parms_d
        }, ckpt_fp)

    def forward(self, **kwargs):
        """return output"""
        raise NotImplementedError
    
    def train_step(self, x, y):
        pred = self(x)
        return {"loss": self.loss_fn(pred, y)}

    def evaluate_step(self, x, y):
        pred = self(x)
        pred = torch.max(pred, dim=-1)[1]
        return {"pred": pred, "target": y}
    
    

    
def seq_len_to_mask(seq_len, max_len=None):
    """
    >>> size = torch.randint(3, 10, (3,)) # e.g., [3,6,6]
    >>> seq_len_to_mask(size).shape == torch.Size([3,size.max()])
    True
    >>> seq_len_to_mask(size, 10).shape   # True/False matrix
    torch.Size([3, 10])
    """
    if isinstance(seq_len, np.ndarray):
        assert len(np.shape(seq_len)) == 1, f"seq_len can only have one dimension, got {len(np.shape(seq_len))}."
        max_len = int(max_len) if max_len else int(seq_len.max())
        broad_cast_seq_len = np.tile(np.arange(max_len), (len(seq_len), 1))
        mask = broad_cast_seq_len < seq_len.reshape(-1, 1)

    elif isinstance(seq_len, torch.Tensor):
        assert seq_len.dim() == 1, f"seq_len can only have one dimension, got {seq_len.dim() == 1}."
        batch_size = seq_len.size(0)
        max_len = int(max_len) if max_len else seq_len.max().long()
        broad_cast_seq_len = torch.arange(max_len).expand(batch_size, -1).to(seq_len)
        mask = broad_cast_seq_len.lt(seq_len.unsqueeze(1))
        
    else:
        raise TypeError("Only support 1-d numpy.ndarray or 1-d torch.Tensor.")

    return mask

    
def xpad(arr, *shape, pad_value=0, dtype=float, rtn_type='numpy'):
    def helper(arr, *shape):
        if not shape: return 
        if len(shape) == 1: return np.array(arr, dtype=dtype)
        _arr = np.full(shape, fill_value=pad_value, dtype=dtype)
        for i, x in enumerate(arr):
            if isinstance(x, np.ndarray):
                size = min(shape[1], len(x))
                _arr[i, :size] = x[:size]
            else:
                rtn = helper(x, *shape[1:])
                _arr[i, :len(rtn)] = rtn
        return _arr
    if not shape:
        if hasattr(arr, 'shape'): shape = arr.shape
        else: shape = [len(arr)]
    out = helper(arr, *shape)
    if rtn_type == 'tensor':
        return torch.from_numpy(out)
    return out

def xmove(args, device):
    if not torch.cuda.is_available() or device is None:
        return
    if isinstance(args, list):
        for arg in args: xmove(arg, device)
    elif isinstance(args, dict):
        for key, value in args.items():
            if isinstance(value, torch.Tensor):
                args[key] = value.to(device)
    else:
        raise TypeError("only dictionary inputs are supported, please change your collate function")

def xgroup(iterable, ndigits=None):
    from collections import defaultdict
    def rd(num, digit=None):
        if digit: num = round(num, digit)
        return num
    out = defaultdict(dict)
    for key in iterable:
        if '|' in key:
            left,right = key.rsplit('|',1)
            out[right][left] = rd(iterable[key], ndigits)
            if '|' in left:
                out[right] = xgroup(out[right])
        else:
            out[key] = rd(iterable[key], ndigits)
    return out

class MyLoss(nn.Module):
    def __init__(self, samples_per_cls=[10,1], no_of_classes=2, loss_type='focal', re_weight=False):
        super(MyLoss, self).__init__()
        self.loss_type = loss_type
        beta, self.gamma = 0.9999, 2.0
        if re_weight:
            if not samples_per_cls:
                samples_per_cls = [1] * no_of_classes
            samples_per_cls = torch.tensor(samples_per_cls)
            effective_num = 1.0 - torch.pow(beta, samples_per_cls)
            weights = (1.0 - beta) / effective_num
            self.weights = nn.Parameter(weights.softmax(-1) * no_of_classes, requires_grad=False)
        else:
            self.weights = None
        self.n_cls = no_of_classes

    def forward(self, pred, target, seq_len=None, is_training=True):
        n = len(pred)
        # 1. prepare weight: weight_ (batch-size, n-class)
        if self.weights is not None:
            weight_ = self.weights.expand(n, self.n_cls)       
            weight_ = weight_[torch.arange(n), target]
            weight_ = weight_[:,None].repeat(1, self.n_cls)
        else:
            weight_ = None
        # 2. prepare inputs (label has to be fload!)
        target_ = F.one_hot(target, self.n_cls).float()
        loss_type, gamma = self.loss_type, self.gamma
        if loss_type == "focal":
            loss = focal_loss(target_, pred, weight_, gamma)
        elif loss_type == "sigmoid":
            loss = F.binary_cross_entropy_with_logits(pred,target_,weight_)
        elif loss_type == "softmax":
            pred = pred.softmax(dim = 1)
            loss = F.binary_cross_entropy(pred, target_, weight_)
        return loss



def focal_loss(labels, logits, alpha, gamma):
    """Compute the focal loss between `logits` and `labels`
    """    
    BCLoss = F.binary_cross_entropy_with_logits(logits, labels, reduction = "none")

    if gamma == 0.0:
        modulator = 1.0
    else:
        modulator = torch.exp(-gamma * labels * logits - gamma * torch.log(1 + 
            torch.exp(-1.0 * logits)))
             
    weighted_loss = alpha * modulator * BCLoss
    focal_loss = weighted_loss.sum()/labels.sum()
    return focal_loss


def f1_loss(y_true:torch.Tensor, y_pred:torch.Tensor, is_training=False, epsilon=1e-7) -> torch.Tensor:
    """Compute the f1 loss between `logits` and `labels`
    """
    if y_pred.ndim == 2: y_pred = y_pred.argmax(dim=1)
    tp = (y_true * y_pred).sum().to(torch.float32)
    fp = ((1 - y_true) * y_pred).sum().to(torch.float32)
    fn = (y_true * (1 - y_pred)).sum().to(torch.float32)
    precision = tp / (tp + fp + epsilon)
    recall = tp / (tp + fn + epsilon)
    f1 = 2* (precision*recall) / (precision + recall + epsilon)
    f1.requires_grad = is_training
    return f1

    
def make_fc(input_dim, hidden_sizes):
    r"""Returns a ModuleList of fully connected layers. 
    
    .. note::
    
        All submodules can be automatically tracked because it uses nn.ModuleList. One can
        use this function to generate parameters in :class:`BaseNetwork`. 
    
    Example::
    
        >>> make_fc(3, [4, 5, 6])
        ModuleList(
          (0): Linear(in_features=3, out_features=4, bias=True)
          (1): Linear(in_features=4, out_features=5, bias=True)
          (2): Linear(in_features=5, out_features=6, bias=True)
        )
    
    Args:
        input_dim (int): input dimension in the first fully connected layer. 
        hidden_sizes (list): a list of hidden sizes, each for one fully connected layer. 
    
    Returns:
        nn.ModuleList: A ModuleList of fully connected layers.     
    """
    assert isinstance(hidden_sizes, list), f'expected list, got {type(hidden_sizes)}'
    
    hidden_sizes = [input_dim] + hidden_sizes
    
    fc = []
    for in_features, out_features in zip(hidden_sizes[:-1], hidden_sizes[1:]):
        fc.append(nn.Linear(in_features=in_features, out_features=out_features))
    
    fc = nn.ModuleList(fc)
    
    return fc

def make_dnn(input_dim: int, hidden_sizes: List[int], hidden_units: List[str] = 'relu', dropouts: List[float] = 0.0, bias=True, init_weight=True, last_act=False):
    return DNN(
        input_dim, hidden_sizes, hidden_units, dropouts, bias, init_weight, last_act
    )

class DNN(nn.Sequential):
    def __init__(
        self, input_dim: int, hidden_sizes: List[int], hidden_units: List[str] = 'relu', dropouts: List[float] = 0.0,  bias=True, init_weight=True, last_act='identity'
    ):
        n = len(hidden_sizes)
        hidden_sizes = self.check_list(hidden_sizes, n)
        hidden_units = self.activate(*(
            self.check_list(hidden_units, n-1) + self.check_list(last_act)
        ))
        dropouts = self.check_list(dropouts, n)
        hidden_sizes = [input_dim] + hidden_sizes
        fc = []
        for i,(in_dim, out_dim) in enumerate(zip(hidden_sizes[:-1], hidden_sizes[1:])):
            fc.extend([
                nn.Dropout(p=dropouts[i]),
                nn.Linear(
                    in_features=in_dim, out_features=out_dim, bias=bias
                ),
                hidden_units[i]
            ])
        super(DNN, self).__init__(*fc)
        if init_weight: self.apply(make_weight)

    @property
    def output_dim(self):
        return self[-2].out_features

    @staticmethod
    def activate(*units):
        def f(unit):
            if unit == 'relu': return nn.ReLU(True)
            elif unit == 'prelu': return nn.PReLU()
            elif unit == 'lrelu': return nn.LeakyReLU(0.2, True)
            elif unit == 'tanh': return nn.Tanh()
            elif unit == 'sigmoid': return nn.Sigmoid()
            elif unit == 'identity': return nn.Identity()
        return list(map(f, filter(None, units)))

    @staticmethod
    def check_list(x, n=1):
        if not isinstance(x, (tuple, list)):
            x = [x] * n
        return x[:n]
        

def clip_weight(layer, min_value=0.0, max_value=1.0):
    def func(m):
        cls = m.__class__.__name__
        if cls.find('Linear') != -1:
            with torch.no_grad():
                m.weight.clamp_(min_value, max_value)
    layer.apply(func)

def make_weight(m):
    cls = m.__class__.__name__
    if cls.find('Conv') != -1:
        nn.init.normal_(m.weight, 0.0, 0.02)
    elif cls.find('BatchNorm') != -1:
        nn.init.normal_(m.weight, 1.0, 0.02)
        if m.bias is not None: nn.init.zeros_(m.bias)
    elif cls.find('Linear') != -1:
        nn.init.kaiming_uniform_(m.weight, a=1)
        if m.bias is not None: nn.init.constant_(m.bias, 0)


def make_cnn(input_channel, channels, kernels, strides, paddings):
    r"""Returns a ModuleList of 2D convolution layers. 
    
    .. note::
    
        All submodules can be automatically tracked because it uses nn.ModuleList. One can
        use this function to generate parameters in :class:`BaseNetwork`. 
        
    Example::
    
        >>> make_cnn(input_channel=3, channels=[16, 32], kernels=[4, 3], strides=[2, 1], paddings=[1, 0])
        ModuleList(
          (0): Conv2d(3, 16, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1))
          (1): Conv2d(16, 32, kernel_size=(3, 3), stride=(1, 1))
        )
    
    Args:
        input_channel (int): input channel in the first convolution layer. 
        channels (list): a list of channels, each for one convolution layer.
        kernels (list): a list of kernels, each for one convolution layer.
        strides (list): a list of strides, each for one convolution layer. 
        paddings (list): a list of paddings, each for one convolution layer. 
    
    Returns:
        nn.ModuleList: A ModuleList of 2D convolution layers.
    """
    N = len(channels)
    
    for item in [channels, kernels, strides, paddings]:
        assert isinstance(item, list), f'expected as list, got {type(item)}'
        assert len(item) == N, f'expected length {N}, got {len(item)}'
    
    channels = [input_channel] + channels
    
    cnn = []
    for i in range(N):
        cnn.append(nn.Conv2d(in_channels=channels[i], 
                             out_channels=channels[i+1], 
                             kernel_size=kernels[i], 
                             stride=strides[i], 
                             padding=paddings[i], 
                             dilation=1, 
                             groups=1))
    
    cnn = nn.ModuleList(cnn)
    
    return cnn


def make_transposed_cnn(input_channel, channels, kernels, strides, paddings, output_paddings):
    r"""Returns a ModuleList of 2D transposed convolution layers. 
    
    .. note::
    
        All submodules can be automatically tracked because it uses nn.ModuleList. One can
        use this function to generate parameters in :class:`BaseNetwork`. 
        
    Example::
    
        make_transposed_cnn(input_channel=3, 
                            channels=[16, 32], 
                            kernels=[4, 3], 
                            strides=[2, 1], 
                            paddings=[1, 0], 
                            output_paddings=[1, 0])
        ModuleList(
          (0): ConvTranspose2d(3, 16, kernel_size=(4, 4), stride=(2, 2), padding=(1, 1), output_padding=(1, 1))
          (1): ConvTranspose2d(16, 32, kernel_size=(3, 3), stride=(1, 1))
        )
    
    Args:
        input_channel (int): input channel in the first transposed convolution layer. 
        channels (list): a list of channels, each for one transposed convolution layer.
        kernels (list): a list of kernels, each for one transposed convolution layer.
        strides (list): a list of strides, each for one transposed convolution layer. 
        paddings (list): a list of paddings, each for one transposed convolution layer. 
        output_paddings (list): a list of output paddings, each for one transposed convolution layer. 
    
    Returns:
        nn.ModuleList: A ModuleList of 2D transposed convolution layers.
    """
    N = len(channels)
    
    for item in [channels, kernels, strides, paddings, output_paddings]:
        assert isinstance(item, list), f'expected as list, got {type(item)}'
        assert len(item) == N, f'expected length {N}, got {len(item)}'
    
    channels = [input_channel] + channels
    
    transposed_cnn = []
    for i in range(N):
        transposed_cnn.append(nn.ConvTranspose2d(in_channels=channels[i], 
                                                 out_channels=channels[i+1], 
                                                 kernel_size=kernels[i], 
                                                 stride=strides[i], 
                                                 padding=paddings[i], 
                                                 output_padding=output_paddings[i],
                                                 dilation=1, 
                                                 groups=1))
    
    transposed_cnn = nn.ModuleList(transposed_cnn)
    
    return transposed_cnn


from shutil import rmtree, move, copyfile
from pathlib import Path
import inspect, optuna, os
from typing import Callable, Dict, Mapping
from optuna.pruners import MedianPruner, HyperbandPruner, NopPruner
from optuna.samplers import PartialFixedSampler, RandomSampler
from optuna._transform import _SearchSpaceTransform
from concurrent.futures import ProcessPoolExecutor
import multiprocessing as mp
from tqdm import tqdm
import numpy as np
import pandas as pd

import torch

from tknlp import Args, color_str, ojoin, oexists
from tknlp.utils import timeit, CloudpickleWrapper

from hydra.core.override_parser.types import (
    Sweep, RangeSweep, ChoiceSweep, IntervalSweep, Transformer, QuotedString
)
from optuna.distributions import (
    CategoricalDistribution, IntLogUniformDistribution, IntUniformDistribution, LogUniformDistribution, UniformDistribution, BaseDistribution, DiscreteUniformDistribution
)
from fastNLP.envs import get_global_rank, is_cur_env_distributed
from copy import copy
from random import random, shuffle
from itertools import product
from functools import reduce

@timeit(color='green', bold=True)
def run_multiple(run, config, seeds, max_workers, chunksize=1, use_gpu=False, gpu_ids=-1, copy_file=False, prune_exp=False, level=2, remove_logs=True, remove_db=True):
    r"""A convenient function to parallelize the experiment (master-worker pipeline). 
    
    It is implemented by using `concurrent.futures.ProcessPoolExecutor`
                
    Args:
        run (function): a function that defines an algorithm, it must take the 
            arguments `(config, seed, device, logdir)`
        config (args): a :class:`Args` object defining all configuration settings parsed via Hydra
        seeds (list): a list of random seeds
        max_workers (int): argument for ProcessPoolExecutor. if `None`, then all experiments run serially.
        chunksize (int): argument for Executor.map()
        use_gpu (bool): if `True`, then use CUDA. Otherwise, use CPU.
        gpu_ids (list): if `None`, then use all available GPUs. Otherwise, only use the
            GPU device defined in the list. 
        level (int): argparse level, default to be able to parse A.B.C
    """
    check_valid_config(config)
    checkpoint = ojoin(config.checkpoint, config.exp_name)
    logpath = ojoin(config.logpath, config.exp_name)
    print(color_str(f'\nExperiment starts. Loggings are stored in {checkpoint}. ', 'cyan', bold=True))

    # recognize distributions in configurations
    parsed = config.parse_args()
    sweep_config = SweepConfig(parsed)
    configs: list = sweep_config.make_config(checkpoint=checkpoint, logpath=logpath)
    KEYS: dict = sweep_config.search_space
    
    # create logging and model checkpoint dir -- move to /tmp if necessary (ask yes?)
    prepare_path(logpath, remove_exists=remove_logs)
    prepare_path(checkpoint, remove_exists=remove_db)
    # save .py files from source folder
    # prepare_copy(copy_file, run, checkpoint)
    write_analysis(
        ojoin(checkpoint, 'analysis.py'), 
        checkpoint, config.exp_name, config.optim_metric
    )
    
    # Create unique id for each job
    jobs = list(enumerate(product(configs, seeds), start=1))
    KEYS['seed'] = seeds
    storage = f"sqlite:///{str(checkpoint)}/experiment.db"
    pruner = HyperbandPruner() if prune_exp else NopPruner()
    study = optuna.create_study(
        study_name=config.exp_name,
        pruner=pruner,
        direction=config.optim_direction,
        storage=storage,
        load_if_exists = True
    )
    for param, space in KEYS.items(): # save config in database
        study._storage.set_study_user_attr(study._study_id, param, space)
    
    # define how each job should be done, call (run) with parameter grid
    def _run(job):
        job_id, (config, seed) = job
        # VERY IMPORTANT TO AVOID GETTING STUCK, oversubscription
        # see following links
        # https://github.com/pytorch/pytorch/issues/19163
        # https://software.intel.com/en-us/intel-threading-building-blocks-openmp-or-native-threads
        torch.set_num_threads(1)
        if use_gpu == True:
            num_gpu = torch.cuda.device_count()
            if gpu_ids is None or gpu_ids == -1:  # use all GPUs
                device_id = job_id % num_gpu
            elif isinstance(gpu_ids, int):
                device_id = gpu_ids
            elif isinstance(gpu_ids, list):
                assert all([i >= 0 and i < num_gpu for i in gpu_ids])
                device_id = gpu_ids[job_id % len(gpu_ids)]
            torch.cuda.set_device(device_id)
            device = torch.device(f'cuda:{device_id}')
        else:
            device = torch.device('cpu')
            
        print('\n\n')
        print(f'@ Experiment: ID: {config["exp_id"]} ({len(configs)}), Seed: {seed}, Device: {device}, Job: {job_id} ({len(jobs)}), PID: {os.getpid()}')
        print('#'*80)
        # config.pop('_post_init_', True) # config is already well parsed no need to do it again
        _args = Args(**config, device=device, seed=seed, level=level)
        _args.set('task_name', ' | '.join(f'{k}={_args.search(k)}' for k in ('exp_name', 'seed', *KEYS)))
        pruner = HyperbandPruner() if prune_exp else NopPruner()
        study = optuna.load_study(study_name=_args.exp_name, storage=storage, pruner=pruner)

        def get_value(trial_id):
            STATE = optuna.trial.TrialState
            trial = study._storage.get_trial(trial_id)
            if trial.state == STATE.COMPLETE:
                return {'metric': trial.values[0]} #in case it's multi-objective
            else:
                return {'metric': 0} #issue: NA might be more appropriate

        def objective(trial):
            for k, vs in KEYS.items():
                trial._suggest(k, sweep_config.dist.get(k, CategoricalDistribution(vs)))
            _args.update(trial = trial, trial_id = trial._trial_id, pid=os.getpid())
            result = run(_args)
            if use_gpu: torch.cuda.empty_cache()
            return result

        fixed_params = {k: _args.search(k) for k in KEYS}
        study.sampler = PartialFixedSampler(fixed_params, study.sampler)
        study.optimize(objective, n_trials=1, gc_after_trial=True)
        # since we have only 1 trial to run, we can index it with -1
        return {'exp_id': _args.exp_id, 'trial_id': _args.trial_id, **fixed_params, **get_value(_args.trial_id)}
    
    if (max_workers is None) or (isinstance(gpu_ids, list) and len(gpu_ids)==1) or (isinstance(gpu_ids, int) and gpu_ids!=-1):
        results = [_run(job) for job in jobs]
    else:
        with ProcessPoolExecutor(max_workers=min(max_workers, len(jobs))) as executor:
            n = len(jobs)
            results = list(tqdm(executor.map(CloudpickleWrapper(_run), jobs, chunksize=chunksize), total=n))
    print(color_str(f'\nExperiment finished. Loggings are stored in {checkpoint}. ', 'cyan', bold=True))
    result_d = pd.DataFrame(results)
    with open(ojoin(checkpoint, 'result.tsv'), 'w') as fw:
        fw.write(repr(result_d))
    return results

class SweepConfig:

    def __init__(self, config_d: Dict, n_repeat: int = 1):
        self.config_d = config_d
        self.dist, self.search_space = {}, {}
        self.n_repeat = n_repeat
        self.sampler = self.prepare_sampler()

    def prepare_sampler(self):
        _rng = np.random.RandomState(seed=1000)
        def sample_independent(key, dist):
            trans = _SearchSpaceTransform({key: dist})
            trans_params = _rng.uniform(trans.bounds[:, 0], trans.bounds[:, 1])
            return trans.untransform(trans_params)[key]
        return sample_independent

    @property
    def n_trials(self):
        return reduce(lambda x, y: x*y, map(len, self.search_space.values()))

    def make_config(self, **fixed_dict):
        lst = self.make_config_helper(**{**self.config_d, **fixed_dict})
        lst *= self.n_repeat
        return [{'exp_id': i+1, **d} for i,d in enumerate(lst)]
    
    def make_config_helper(self, **d):
        sdist, sspace, fix = {}, {}, {}
        for k, v in d.items():
            if isinstance(v, Sweep):
                self.dist[k] = sdist[k] = self.parse_sweep(v)
                self.search_space[k] = sspace[k] = self.parse_dist(k, sdist[k])
            elif isinstance(v, QuotedString):
                fix[k] = v.text
            elif isinstance(v, (Dict, Args)):
                sspace[k] = self.make_config_helper(**v)
            else:
                fix[k] = v
        keys, values = sspace.keys(), sspace.values()
        grid_product = list(dict(zip(keys, vs)) for vs in product(*values))
        for config in grid_product:
            for k,v in config.items():
                if isinstance(v, str) and v.startswith('sample@'):
                    param = v.lstrip('sample@')
                    config[k] = self.sampler(k, sdist[param])
        # n_trials = reduce(lambda x, y: x*y, map(len, sspace.values()))
        if sspace:
            return [{**fix, **g} for g in grid_product]
        else:
            return [{**fix}]
                
    
    @staticmethod
    def parse_dist(key, X):
        if isinstance(X, CategoricalDistribution):
            return X.choices
        elif isinstance(X, IntUniformDistribution):
            return [X.low + i*X.step for i in range((X.high - X.low) // X.step)]
        elif isinstance(X, DiscreteUniformDistribution):
            return [X.low + i * X.q for i in range(int((X.high - X.low) // X.q))]
        else:
            return [f'sample@{key}']    # all other distribution is 

    @staticmethod
    def parse_sweep(X):
        transform = Transformer.encode
        def sort_X(start, end, step=None):
            if start > end:
                return end, start, -step
            return start, end, step

        if isinstance(X, ChoiceSweep):
            lst = copy(X.list)
            X.shuffle and shuffle(lst)
            return CategoricalDistribution([x for x in map(transform, lst)])

        elif isinstance(X, RangeSweep):
            if X.shuffle:
                lst = list(X.range()); shuffle(lst)
                return CategoricalDistribution([x for x in map(transform, lst)])
            start, stop, step = sort_X(X.start, X.stop, X.step)
            if (isinstance(start, float) or isinstance(stop, float) or isinstance(step, float)):
                return DiscreteUniformDistribution(start, stop, step)
            start, stop, step = map(int, (start, stop, step))
            return IntUniformDistribution(start, stop, step)

        elif isinstance(X, IntervalSweep):
            tags, start, end, _ = X.tags, *sort_X(X.start, X.end)
            if "log" in tags:
                if isinstance(start, int) and isinstance(end, int):
                    return IntLogUniformDistribution(int(start), int(end))
                return LogUniformDistribution(start, end)
            else:
                if isinstance(start, int) and isinstance(end, int):
                    return IntUniformDistribution(start, end)
                return UniformDistribution(start, end)

          
def clear_path(path):
    (sp := path/'source_files').exists() and rmtree(sp)
    (sp := path/'experiment.db').exists() and sp.unlink()
    (sp := path/'analysis.py').exists() and sp.unlink()
    (sp := path/'logs').exists() and rmtree(sp)
    (sp := path/'checkpoint').exists() and rmtree(sp)

    
def prepare_path_helper(path):
    def pmove(source, target):
        if oexists(target): rmtree(target)
        move(source, target)
    import tempfile
    NAME = os.path.basename(path)
    temp_dir = ojoin(tempfile.gettempdir(), NAME, 'logs', create_if_not_exist=True)
    pmove(path, temp_dir)
    print(color_str(f'\nSame exp_name={NAME} detected, directory is moved to {temp_dir}. ', 'cyan', bold=True))

def prepare_path(*path, remove_exists=True):
    for p in path: 
        if remove_exists and oexists(p):
            prepare_path_helper(p)
        os.makedirs(p, exist_ok=True)

    
def prepare_copy(run_copy: bool, program: Callable, log_path: Path):
    if run_copy and not check_in_ipython():
        source_path = Path(inspect.getsourcefile(program)).parent
        target_path = Path(log_path)/'source_files'
        copy_recursive(source_path, target_path)

class Grader:
    def __init__(self, metric_name, best_value=0, best_param = {}):
        self.name = metric_name
        self.best_value = best_value
        self.best_param = best_param
    def update(self, value, param):
        self.best_value = value
        self.best_param = param

def copy_recursive(source: Path, target: Path): 
    target.mkdir(exist_ok=True, parents=True)
    for item in source.iterdir():
        if item.is_dir() and not item.name.startswith('__'):
            new_target = target/item.name
            new_target.mkdir(exist_ok=True)
            copy_recursive(item, new_target)
        else:
            for file in source.glob('*.py'):
                copyfile(file, target/file.name)

def check_in_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def check_valid_config(config):
    assert isinstance(config, Args), "you have to instantiate config with dpvs.config.Args class"
    assert 'exp_name' in config, "you have to specify an experiment name"
    assert 'optim_metric' in config, "you have to specify metric key used for model comparisons"
    assert 'optim_direction' in config, "you have to specify direction to which better result is defined"
    assert 'checkpoint' in config, "you have to specify where output/logging is saved"


def write_analysis(path, logpath, name, metric_key):
    from inspect import cleandoc
    py_str = f"""
from tknlp.utils import OptunaAnalysis
study = OptunaAnalysis('{name}', 'sqlite:///{logpath}/{name}.db')
study.df
study.make_plot(category='seed', column='', y='{metric_key}')
study.make_coordinate()
    """
    with open(path, 'w') as f:
        f.write(cleandoc(py_str))


from pathlib import Path
from operator import itemgetter, attrgetter
from collections import OrderedDict
from .utils import oexists, omake, color_str
from .typing_ import Mapping, List

import cloudpickle
import pickle
import yaml
import os
import logging
import numpy as np 
import pandas as pd

os.environ['ANSI_COLORS_DISABLED'] = "True"
from wasabi import Printer

def get_num_gpu():
    with os.popen('nvidia-smi --list-gpus | wc -l') as proc:
        output = proc.read()
    return int(output)

def is_cur_env_distributed():
    return "FASTNLP_GLOBAL_RANK" in os.environ

class Logger:
    _name,_path,_logger = None, None, {}

    @staticmethod
    def get(**kwargs):
        return get_logger(name=kwargs.pop('name', Logger._name), **kwargs)

    @staticmethod
    def clear():
        Logger._name = None

    def __init__(self, name=None, message_only=True, formatter=None, level='INFO', path=''):
        self.log = get_logger(name, message_only, formatter, level, path)
        if is_cur_env_distributed() and os.getenv('FASTNLP_GLOBAL_RANK') != '0':
            self.log.setLevel(logging.ERROR)
            print("Experiment logging is saved in %s"%(color_str(path)))
        else:
            print("Experiment logging is saved in %s"%(color_str(path)))
        Logger._name = name
        Logger._path = path
        Logger.get(name='fastNLP').setLevel('WARNING')

    def info(self, msg, *args, **kwargs):
        self.log.info( msg, *args, **kwargs)

def get_logger(name=None, message_only=True, formatter=None, level='INFO', path=''):
    loggers = Logger._logger
    if name in loggers: return loggers.get(name)
    logger = logging.getLogger(name or __name__)
    logger.setLevel(logging._nameToLevel[level])
    if message_only:
        if formatter is not None:
            formatter = logging.Formatter(formatter)
        else:
            formatter = logging.Formatter('%(name)s :: %(asctime)s ::  %(message)s', "%H:%M")
        if path:
            log = logging.FileHandler(path,mode='w')
        else:
            log = logging.StreamHandler()
        log.setFormatter(formatter)
        logger.addHandler(log)
        # propagate=False: disable message spawned by child logger pass to parent logger
        logger.propagate = False        
        loggers[name] = logger
    return logger

log = Logger.get()

class Recorder:
    r"""Log the information in a dictionary. 
    
    If a key is logged more than once, then the new value will be appended to a list. 
    
    .. note::
    
        It uses pickle to serialize the data. Empirically, ``pickle`` is 2x faster than ``numpy.save``
        and other alternatives like ``yaml`` is too slow and ``JSON`` does not support numpy array. 
    
    .. warning::
    
        It is discouraged to store hierarchical structure, e.g. list of dict of list of ndarray.
        Because pickling such complex and large data structure is extremely slow. Put dictionary
        only at the topmost level. Large numpy array should be saved separately. 
    
    Example:
    
    * Default::
    
        >>> logger = Recorder()
        >>> logger('iteration', 1)
        >>> logger('train_loss', 0.12)
        >>> logger('iteration', 2)
        >>> logger('train_loss', 0.11)
        >>> logger('iteration', 3)
        >>> logger('train_loss', 0.09)
        
        >>> logger
        OrderedDict([('iteration', [1, 2, 3]), ('train_loss', [0.12, 0.11, 0.09])])
        
        >>> if i == 0 or (i+1) % self.config.log_freq == 0:
        >>>     logger.dump(keys=None, index=-1, indent=0, border='-'*50)
        Iteration: [1, 2, 3]
        Train Loss: [0.12, 0.11, 0.09]
        
    * With indentation::
    
        >>> logger.dump(indent=1)
            Iteration: [1, 2, 3]
            Train Loss: [0.12, 0.11, 0.09]
        
    * With specific keys::
    
        >>> logger.dump(keys=['iteration'])
        Iteration: [1, 2, 3]
        
    * With specific index::
    
        >>> logger.dump(index=0)
        Iteration: 1
        Train Loss: 0.12
        
    * With specific list of indices::
    
        >>> logger.dump(index=[0, 2])
        Iteration: [1, 3]
        Train Loss: [0.12, 0.09]
    
    """
    def __init__(self):
        # TODO: wait for popularity of Python 3.7 which dict preserves the order, then drop OrderedDict()
        self.logs = OrderedDict()

    def average(self, key=None):
        if key: 
            return np.mean(self.logs[key])
        else:
            for key in self.logs:
                self.logs[key] = [self.average(key)]

    @property
    def KEYS(self):
        # return set(x.split('_')[-1] for x in self.logs.keys() if x != 'epoch')
        return set(x for x in self.logs.keys() if x != 'epoch')

    def __iter__(self):
        for key in self.KEYS:
            yield key, self.logs[key]
        
    def __call__(self, key, value):
        r"""Log the information with given key and value. 
        
        .. note::
        
            The key should be semantic and each word is separated by ``_``. 
        
        Args:
            key (str): key of the information
            value (object): value to be logged
        """
        if key not in self.logs:
            self.logs[key] = []
        
        self.logs[key].append(value)

    def __getitem__(self, key):
        return self.logs[key]
        
    def dump(self, keys=None, index=None, indent=0, border=''):
        r"""Dump the loggings to the screen.
        
        Args:
            keys (list, optional): a list of selected keys. If ``None``, then use all keys. Default: ``None``
            index (int/list, optional): the index of logged values. It has following use cases:
                
                - ``scalar``: a specific index. If ``-1``, then use last element.
                - ``list``: a list of indicies. 
                - ``None``: all indicies.
                
            indent (int, optional): the number of tab indentation. Default: ``0``
            border (str, optional): the string to print as header and footer
        """
        if keys is None:
            keys = list(self.logs.keys())
        assert isinstance(keys, list), f'expected list, got {type(keys)}'
        
        if index is None:
            index = 'all'
        
        print(border)
        for key in keys:
            if indent > 0:
                print('\t'*indent, end='')  # do not create a new line
            
            if index == 'all':
                value = self.logs[key]
            elif isinstance(index, int):
                value = self.logs[key][index]
            elif isinstance(index, list):
                value = list(itemgetter(*index)(self.logs[key]))
            
            # Polish key string
            key = key.strip().replace('_', ' ').title()
            
            print(f'{key}: {value}')
        print(border)

    def save(self, f):
        r"""Save loggings to a file. 
        
        Args:
            f (str): file path
        """
        with open(f, 'wb') as file:
            if f.endswith('pkl'):
                pickle.dump(self.logs, file)
            elif f.endswith('csv'):
                pd.DataFrame.from_dict(self.logs).to_csv(f, index=False)
            else:
                print(self.logs, file=file)
        
    def clear(self):
        r"""Remove all loggings in the dictionary. """
        self.logs.clear()
        
    def __repr__(self):
        return repr(self.logs)

    def __str__(self):
        return ','.join(set(x.split('_')[-1] for x in self.logs.keys()))

def pickle_load(f, ext='.pkl', encode="ASCII"):
    r"""Read a pickled data from a file. 
    """
    if not f: raise FileNotFoundError()
    if isinstance(f, Path):
        f = f.as_posix() 
        f += [ext,''][f[-4:]==ext]

    with open(f, 'rb') as file:
        return cloudpickle.load(file, encoding=encode)

def pickle_loads(*f):
    for sf in f: 
        yield pickle_load(sf)

def pickle_dump(obj, f, ext='.pkl'):
    r"""Serialize an object using pickling and save in a file. 
    
    Args:
        obj (object): a serializable object
        f (str/Path): file path
        ext (str, optional): file extension. Default: .pkl
    """
    if isinstance(f, Path):
        f = f.as_posix()
        f += [ext,''][f[-4:]==ext]
    
    with open(f, 'wb') as file:
        return cloudpickle.dump(obj=obj, file=file, protocol=pickle.HIGHEST_PROTOCOL)


def yaml_load(f):
    r"""Read the data from a YAML file. 
    """
    if isinstance(f, Path):
        f = f.as_posix()
    
    with open(f, 'r') as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def yaml_dump(obj, f, ext='.yml'):
    r"""Serialize a Python object using YAML and save in a file. 
    
    Args:
        obj (object): a serializable object
        f (str/Path): file path
        ext (str, optional): file extension. Default: .yml
        
    """
    if isinstance(f, Path):
        f = f.as_posix()
    with open(f+ext, 'w') as file:
        return yaml.dump(obj, file, sort_keys=False)

def cache_results(*_cache_fp, _refresh=False):
    r""" a decorator to cache function results
    @reference: FastNLP::core::utils
    :param str `_cache_fp`:     where to read the cache from
    :param bool `_refresh`:     whether to regenerate cache
    """
    def wrapper_(func):
        def wrapper(*args, **kwargs):
            refresh = kwargs.pop('_refresh', _refresh)
            cache_fp = kwargs.pop('_cache_fp', _cache_fp)
            if kwargs.pop('_skipit', False):
                return func(*args, **kwargs)
            if isinstance(cache_fp, str): cache_fp = [cache_fp]
            cond = cache_fp and not refresh and oexists(*cache_fp)
            if not cond:
                omake(*cache_fp)
                results = func(*args, **kwargs)
                if not results:
                    raise RuntimeError("The return value is None. Delete the decorator.")
                if len(cache_fp) == 1 and isinstance(results, (Mapping, List)):
                    results = [results]
                for path, result in zip(cache_fp, results):
                    pickle_dump(result, path)
            else:
                for fp in cache_fp: log.info(f'read cache from {color_str(fp)}')
            return map(pickle_load, cache_fp)
        return wrapper
    return wrapper_


class FileLog:
    def __init__(self, path, overwrite=False):
        self.path = path
        self.msg = Printer(no_print=True, pretty=True)
        overwrite and open(path, 'w').close()
    def __call__(self, msg, mode='a'):
        with open(self.path, mode) as f:
            print(msg, file=f)
    def divider(self, msg, **kwargs):
        self(self.msg.divider(msg, **kwargs))
    def seperate(self):
        self('-'*self.msg.line_max)
    def table(self, data, header=None, divider=True, **kwargs):
        self(self.msg.table(
            data=data, header=header, divider=divider, **kwargs
        )[1:-1])


from typing import List, Union, Literal, Optional, Tuple, Callable, Dict, Any, Iterable, Mapping, Generator, overload, TypeVar
from pathlib import Path
from abc import ABC
from dataclasses import dataclass
from collections import Counter, defaultdict

class CanItemDataType(ABC):
    @classmethod
    def __subclasshook__(cls, subclass: Any) -> Union[bool, Any]:
        if cls is CanItemDataType:
            item = getattr(subclass, 'item', None)
            return callable(item)
        return NotImplemented
    

_T = TypeVar("_T")

IntList = List[int] # A list of token_ids
StrList = List[str] # A list of string

@dataclass
class NerSample:
    input_ids: IntList
    attention_masks: IntList
    labels: IntList
    words: StrList




import sys
import csv
import numpy as np
from dataclasses import dataclass
from torch.utils.data import Dataset, DataLoader
from torch import Tensor, tensor
import torch.nn.functional as F
from typing import Generator, Optional

from fastNLP.core.dataloaders import MixDataLoader, utils
from fastNLP.core.metrics import Metric, ClassifyFPreRecMetric
from torchmetrics.classification.auroc import BinaryAUROC
from torchmetrics import Metric
from typing import Callable
from collections import Counter

csv.field_size_limit(sys.maxsize)


class Concat(MixDataLoader):

    def __init__(self, **dl):
        ds, collate_fn = {}, {}
        for name, loader in dl.items():
            ds[name] = loader.dataset
            collate_fn[name] = collate_helper(loader._get_collator())
        super(Concat, self).__init__(datasets=ds, mode='sequential', collate_fn=collate_fn, batch_size=loader.batch_size, drop_last=loader.drop_last)
    
    
def collate_helper(collate_fn: Callable):
    def collate_batch(batch):
        idx, rtn = utils.indice_collate_wrapper(collate_fn)(batch)
        return rtn
    return collate_batch


class Data:
    keys = ['train', 'val', 'test', 'aug']
    def apply(self, func, **data):
        for mode, old_data in data.items():
            assert mode in self.keys, "%s must be one of %s"%(mode, self.keys)
            new_data = func(old_data or getattr(self, mode))
            if isinstance(new_data, Generator):
                setattr(self, mode, list(zip(*new_data)))
            else:
                setattr(self, mode, new_data)
    def __iter__(self):
        for mode in self.keys: 
            yield getattr(self, mode)
            
@dataclass
class Datasets(Data):
    """Class for keeping track of datasets"""
    train: Optional[Dataset] = None
    test: Optional[Dataset] = None
    val: Optional[Dataset] = None
    aug: Optional[Dataset] = None

@dataclass
class Dataloaders(Data):
    """Class for keeping track of dataloaders"""
    train: DataLoader
    test: DataLoader 
    val: Optional[DataLoader] = None
    aug: Optional[Dataset] = None

class ClassifyMetric(ClassifyFPreRecMetric):

    def __init__(self, only_gross=True, f_type='macro', beta=1, verbose=True, num_class=2, aggregate_when_get_metric=None, **kwargs):
        """
        Args
        ----
        @verbose                control whether including tp/fp/tn/fn
        """
        assert num_class == 2, "It works in two-class classification problem only !!!"
        super(ClassifyMetric, self).__init__(only_gross=only_gross, f_type=f_type, beta=beta, aggregate_when_get_metric=aggregate_when_get_metric)
        self.verbose = verbose

    def get_metric(self) -> dict: 
        result = super(ClassifyMetric, self).get_metric()
        tps, fps = zip(*self.all_gather_object([self._tp, self._fp]))
        tps, fps = sum(tps, Counter()), sum(fps, Counter())
        if self.verbose:
            result.update(
                tp=tps[1], fn=fps[0], fp=fps[1], tn=tps[0], 
                # binary=2*tps[1]/(2*tps[1] + fps[0] + fps[1])
            )
        return result


class ClassifyAUC(BinaryAUROC):

    def update(self, pred: Tensor, target: Tensor) -> None:  # type: ignore
        super(ClassifyAUC, self).update(pred, target)

    def compute(self): # convert tensor to number
        result = super(ClassifyAUC, self).compute()
        return {'auc': result.item()}


class ClassifyFunc(Metric):
    is_differentiable: bool = True
    higher_is_better: bool = False
    full_state_update: bool = False
    sum_measure: Tensor
    total_count: Tensor
    def __init__(self, func: Callable, name: str = '', **kwargs):
        super().__init__(**kwargs)
        self.func = func
        self.name = name or 'measure'
        self.add_state("sum_measure", default=tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total_count", default=tensor(0), dist_reduce_fx="sum")
    
    def update(self, pred: Tensor, target: Tensor) -> None:  # type: ignore
        n = len(target)
        self.sum_measure += self.func(pred.float(), target.float()) * n
        self.total_count += n

    def compute(self) -> Tensor:
        return {self.name: (self.sum_measure / self.total_count).item()}

        
class CrossEntropy(Metric):
    """Torchmetrics cross entropy loss implementation.
    This class implements cross entropy loss as a :class:`torchmetrics.Metric` so that it can be returned by the
    :meth:`~.ComposerModel.metrics`.
    Args:
        ignore_index (int, optional): Specifies a target value that is ignored
            and does not contribute to the input gradient. ``ignore_index`` is only applicable when the target
            contains class indices. Default: ``-100``.
        dist_sync_on_step (bool, optional): sync distributed metrics every step. Default: ``False``.
    """
    full_state_update = False

    def __init__(self, name: str = '', ignore_index: int = -100, aggregate_when_get_metric: bool = False, loss_type='sigmoid'):
        super().__init__(dist_sync_on_step=aggregate_when_get_metric)
        self.ignore_index = ignore_index
        self.name = name or 'loss'
        self.loss_type = loss_type
        self.add_state('sum_loss', default=tensor(0.), dist_reduce_fx='sum')
        self.add_state('total_batches', default=tensor(0), dist_reduce_fx='sum')

    def update(self, logit: Tensor, target: Tensor) -> None:
        """Update the state with new predictions and targets."""
        # Loss calculated over samples/batch, accumulate loss over all batches
        target_ = F.one_hot(target, 2).float()
        if self.loss_type == 'sigmoid':
            self.sum_loss += F.cross_entropy(logit, target, ignore_index=self.ignore_index)
        elif self.loss_type == 'softmax':
            prob = logit.softmax(dim = 1)
            self.sum_loss += F.binary_cross_entropy(prob, target_)
        assert isinstance(self.total_batches, Tensor)
        self.total_batches += 1

    def compute(self) -> Tensor:
        """Aggregate state over all processes and compute the metric."""
        # Return average loss over entire validation dataset
        assert isinstance(self.total_batches, Tensor)
        assert isinstance(self.sum_loss, Tensor)
        return {self.name: (self.sum_loss / self.total_batches).item()}


import os
import sys
import hashlib
import torch
import numpy as np
import random

from datetime import timedelta, datetime
from time import perf_counter
from contextlib import contextmanager
from functools import reduce, wraps
from colorama import Fore, Style
from itertools import zip_longest

from .typing_ import (
    Mapping, Dict, Callable, Union, Path, Generator, defaultdict, Iterable, List, Any
)
from .model import seq_len_to_mask


# -------------------------- General Utility Function ----------------------------------
# seed_all: fix random seet to make experiment replicable
# one_time: set a timer to make sure function is run only one time
# encodes: turn long string into shortened sha256 hexdigits 
# timeit: return running time of callbed function
# color_str: style a string with color

# -------------------------- Xtools on Data Structure ----------------------------------
# flatten_dict: flatten a nested dictionary
# filter_empty: filter out entries in list with rules and indices
# xrank: retrieve {topk} {mode: max/min} element in {dim}, with max {seq_len} on each {dim}
# xlen: infer the shape of a nested list after padding
# xpad: pad array {arr} into a predetermined shape, with {pad_value}
# xmove: move all data in {args} to {device}

# -------------------------- O/I Related Function --------------------------------------
# get_dir: get execute directory of main program
# find_dotenv: find .env file recursively from current to root
# ojoin: take a list of dir-name and concatenate them to make a path 
# omake: create directory for each input arguments
# oexits: do all input paths exist, return False if any input argument isn't a path
# opath: return absolute path
# oreplace: replace filename only in the same directory
# owalk: walk through a directory
# CloudpickleWrapper: serialize/picklize contents


def seed_all(seed: int) -> None:
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    generator = torch.Generator()
    generator.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    
def one_time(func: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        if not wrapper.has_been_run:
            wrapper.has_been_run = True
            return func(*args, **kwargs)
    wrapper.has_been_run = False
    return wrapper

def find_dotenv(path_name: str='', file_name: str='.env') -> Union[None, Path]:
    """find .env in current and all its parent"""
    path = path_name or get_dir(); path = Path(path).absolute()
    while path.as_posix() != path.root:
        attempt = path/file_name
        if attempt.exists(): return attempt
        path = path.parent
    return None

def get_dir() -> Path:
    if check_in_ipython():
        return os.getcwd()
    return odir(sys.argv[0])

def check_in_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def odir(path, has_extension=True):
    _root, _ext = os.path.splitext(path)
    if has_extension and _ext != '':
        _root = os.path.dirname(_root)
    return _root

def find_dotenv(path_name='', file_name='.env'):
    """find .env in current and all its parent"""
    path = path_name or get_dir(); path = Path(path).absolute()
    while path.as_posix() != path.root:
        attempt = path/file_name
        if attempt.exists(): return attempt
        path = path.parent
    return None

def get_dir():
    """get execute directory of main program """
    if check_in_ipython():
        return os.getcwd()
    return odir(sys.argv[0])

def _flatten_dict(dct, prefix=''):
    if isinstance(dct, (Mapping, Dict)):
        if prefix: prefix += '.'
        for k, v in dct.items():
            yield from _flatten_dict(v, prefix+str(k))
    else:
        yield prefix, dct
    
def flatten_dict(dct: Dict, rtn_dct=True) -> Generator:
    if rtn_dct:
        return dict(_flatten_dict(dct))
    else:
        return _flatten_dict(dct)

def filter_empty(*lst, empty=None, remove_index=[]):
    idx, rtn = zip(*filter(lambda c: c[1] != empty and c[0] not in remove_index, enumerate(lst[0])))
    rtn_lst = [list(rtn)]
    for l in lst[1:]: rtn_lst.append([l[i] for i in idx])
    return rtn_lst

def ojoin(*args, create_if_not_exist=False) -> Path:
    path = os.path.join(*args)
    if create_if_not_exist: 
        omake(path)
    return path

def oreplace(path, toreplace: str, create_if_not_exist=False):
    parent = odir(path)
    return ojoin(parent, toreplace, create_if_not_exist=create_if_not_exist)

def owalk(*args, limit=-1, ext=None):
    for path in args:
        for name in sorted(os.listdir(path)):
            if ext and not name.endswith(ext): continue
            yield path, name
            if limit > 0: limit -= 1
            if limit == 0: break 
        else: continue  # only execute if inner loop not break
        break

def omake(*args):
    for path in map(odir, args):
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

def oexists(*path) -> bool:
    def check_exist(path):
        return os.path.exists(path)
    return all(check_exist(p) for p in path)

def opath(path):
    return os.path.abspath(path)    

def encodes(long_str):
    "encode a long string in a consistent and determiistic manner"
    return hashlib.sha256(long_str.encode()).hexdigest()


class CloudpickleWrapper(object):
    r"""Uses cloudpickle to serialize contents (multiprocessing uses pickle by default)
    
    This is useful when passing lambda definition through Process arguments.
    """
    def __init__(self, x):
        self.x = x
        
    def __call__(self, *args, **kwargs):
        return self.x(*args, **kwargs)
    
    def __getattr__(self, name):
        return getattr(self.x, name)
    
    def __getstate__(self):
        import cloudpickle
        return cloudpickle.dumps(self.x)
    
    def __setstate__(self, ob):
        import pickle
        self.x = pickle.loads(ob)
        
def timeit(_func=None, *, color='green', bold=False):
    def decorator_timeit(f):
        r"""Print the runtime of the decorated function. """
        @wraps(f)
        def wrapper_timeit(*args, **kwargs):
            t = perf_counter()
            out = f(*args, **kwargs)
            total_time = timedelta(seconds=round(perf_counter() - t))
            timestamp = datetime.now().isoformat(' ', 'seconds')
            print(color_str(string=f'\nTotal time: {total_time} at {timestamp}', color=color, bold=bold))
            return out
        return wrapper_timeit
    if _func is None:
        return decorator_timeit
    else:
        return decorator_timeit(_func)

def color_str(string, color='cyan', bold=True):
    if isinstance(string, Path): string = string.as_posix()
    elif isinstance(string, (float, int)): string = str(string)
    colors = {'red': Fore.RED, 'green': Fore.GREEN, 'blue': Fore.BLUE, 'cyan': Fore.CYAN, 
              'magenta': Fore.MAGENTA, 'black': Fore.BLACK, 'white': Fore.WHITE}
    style = colors[color]
    if bold:
        style += Style.BRIGHT
    out = style + string + Style.RESET_ALL
    return out

def xrank(tensor, dim=0, topk=0, mode='max', pad=0., seq_len=None):
    if mode == 'max':
        try:
            attn, idx = tensor.max(dim=dim)
        except IndexError:
            pass
        largest, invalid =True, float('-inf')
    else:
        attn, idx = tensor.min(dim=dim)
        largest, invalid =False, float('inf')
    if topk <= 0 or attn.size(-1) <= topk: 
        return attn, idx
    if seq_len is not None:
        mask = seq_len_to_mask(seq_len) 
        attn.masked_fill_(~mask, invalid)
    else:
        mask = torch.ones_like(attn).bool()
    _, i = attn.topk(topk, largest=largest, dim=-1)
    attn.masked_fill_(mask.scatter(-1,i,False), pad)
    return attn, i

def xtable(
    dicts: List[Dict], 
    keys: List[str] = None, 
    pads: List[str] = None, 
    fcodes: List[str] = None, 
    convert_headers: Dict[str, Callable] = None, 
    header_names: List[str] = None, 
    skip_none_lines: bool = False, 
    skip_head_lines: bool = False,
    replace_values: Dict[str, Any] = None, index: List[str] = None
):
    """ Generate ascii table from dictionary
        Copyright: https://stackoverflow.com/questions/40056747/print-a-list-of-dictionaries-in-table-form
    dicts: input dictionary list; empty lists make keys OR header_names mandatory
    keys: order list of keys to generate columns for; no key/dict-key should suffix with '____' else adjust code-suffix
    pads: indicate padding direction and size, eg <10 to right pad alias left-align
    fcodes: formating codes for respective column type, eg .3f
    convert_headers: apply converters(dict) on column keys k, eg timestamps
    header_names: supply for custom column headers instead of keys
    skip_none_lines: skip line if contains None
    replace_values: specify per column keys k a map from seen value to new value;
                    new value must comply with the columns fcode; CAUTION: modifies input (due speed)
    """
    # optional arg prelude
    def infer_type(v):
        if isinstance(v, float): return ('.4f', '>8') 
        elif isinstance(v, int): return ('d', '>6')
        else: return ('s', '<8')
    if index is not None:
        dicts = [{'index': y, **x} for x,y in zip(dicts, index)]
    if keys is None:
        if len(dicts) > 0:
            keys = dicts[0].keys()
        elif header_names is not None:
            keys = header_names
        else:
            raise ValueError('keys or header_names mandatory on empty input list')
    if pads is None  and fcodes is None:
        fcodes, pads = zip(*map(infer_type, dicts[0].values()))
    N_keys, N_pads, N_codes = len(keys), len(pads), len(fcodes)
    if N_pads != N_keys:
        raise ValueError(f'bad pad length {len(pads)}, expected: {N_keys}')
    elif N_codes != N_keys:
        raise ValueError(f'bad fcodes length {len(fcodes)}, expected: {N_keys}')
    if convert_headers is None: convert_headers = {}
    if header_names is None: header_names = keys
    if replace_values is None: replace_values = {}
    # build header
    headline = '│'.join(f"{v:{pad}}" for v, pad in zip_longest(header_names, pads))
    underline = '─' * len(headline)
    # suffix special keys to apply converters to later on
    marked_keys = [h + '____' if h in convert_headers else h for h in keys]
    marked_values = {}
    s = '│'.join(f"{{{h}:{pad}{fcode}}}" for h, pad, fcode in zip_longest(marked_keys, pads, fcodes))
    if skip_head_lines:
        lines = []
    else:
        lines = [headline, underline, ]
    for d in dicts:
        none_keys = [k for k, v in d.items() if v is None]
        if skip_none_lines and none_keys:
            continue
        elif replace_values:
            for k in d.keys():
                if k in replace_values and d[k] in replace_values[k]:
                    d[k] = replace_values[k][d[k]]
                if d[k] is None:
                    raise ValueError(f"bad or no mapping for key '{k}' is None. Use skip or change replace mapping.")
        elif none_keys:
            raise ValueError(f'keys {none_keys} are None in {d}. Do skip or use replace mapping.')
        for h in convert_headers:
            if h in keys:
                converter = convert_headers[h]
                marked_values[h + '____'] = converter(d)
        line = s.format(**d, **marked_values)
        lines.append(line)
    return '\n'.join(lines)

def _isnested(arr):
    """whether `arr` is nested iterable object
    >>> _isnested([1,2,3]) #should be False
    """
    try:
        return any(len(x) > 1 for x in arr)
    except TypeError:
        return False
    
def xlen(arr):
    """infer shape of nested list
    >>> lst = [[1,], [2,3], [2]]
    >>> xlen(lst) # [3,2]
    """
    if isinstance(arr, str) or not isinstance(arr, Iterable) or not _isnested(arr):
        return len(arr)
    carr = [xlen(x) for x in arr]    
    if not _isnested(carr):
        return [len(arr), max(carr)]
    else:
        return [len(arr)] + list(map(max, zip_longest(*carr, fillvalue=0)))

def xpad(arr, *shape, pad_value=0, dtype=float, rtn_type='numpy'):
    def helper(arr, *shape):
        if not shape: return 
        if len(shape) == 1: return np.array(arr, dtype=dtype)
        _arr = np.full(shape, fill_value=pad_value, dtype=dtype)
        for i, x in enumerate(arr):
            if isinstance(x, np.ndarray):
                size = min(shape[1], len(x))
                _arr[i, :size] = x[:size]
            else:
                rtn = helper(x, *shape[1:])
                _arr[i, :len(rtn)] = rtn
        return _arr
    if not shape:
        if hasattr(arr, 'shape'): shape = arr.shape
        else: 
            shape = xlen(arr)
    out = helper(arr, *shape)
    if rtn_type == 'tensor':
        return torch.from_numpy(out)
    return out

def xmove(args, device):
    if not torch.cuda.is_available() or device is None:
        return
    if isinstance(args, list):
        for arg in args: xmove(arg, device)
    elif isinstance(args, Mapping):
        for key, value in args.items():
            if isinstance(value, torch.Tensor):
                args[key] = value.to(device)
    else:
        raise TypeError("only dictionary inputs are supported, please change your collate function")

def xgroup(iterable: Iterable, ndigits=None) -> Dict:
    def rd(num, digit=None):
        if digit: num = round(num, digit)
        return num
    out = defaultdict(dict)
    for key in iterable:
        if '|' in key:
            left,right = key.rsplit('|',1)
            out[right][left] = rd(iterable[key], ndigits)
            if '|' in left:
                out[right] = xgroup(out[right])
        else:
            out[key] = rd(iterable[key], ndigits)
    return out


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tknlp']

package_data = \
{'': ['*']}

install_requires = \
['FastNLP==1.0.1',
 'cloudpickle==2.2.0',
 'fitlog==0.9.13',
 'hydra-core==1.2.0',
 'optuna>=3.0.5,<4.0.0',
 'pandas==1.5.1',
 'scikit-learn==1.1.3',
 'scipy==1.8.1',
 'transformers==4.25.1']

setup_kwargs = {
    'name': 'tknlp',
    'version': '0.1.3',
    'description': '',
    'long_description': '# TKNLP\n\n\nRun NLP tools in an easier manner\n\n## Installation\n\n```bash\npip install tknlp\n```\n',
    'author': 'dennislblog',
    'author_email': 'dennisl@udel.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from .utils import one_time, find_dotenv, get_dir, ojoin, oexists
from .cli import Args
from .contribute import color_str, pickle_dump, pickle_load, get_num_gpu
from dotenv import load_dotenv

@one_time
def load_env():
    load_dotenv(find_dotenv(get_dir()))
load_env()


import argparse
import re
import sys
from operator import attrgetter, itemgetter
from pprint import pformat
from copy import deepcopy
from .typing_ import Mapping, Any, Callable, Dict, Path
from .utils import flatten_dict

def search_api(key, obj):
    if key in obj: return obj.get(key)
    for nested in filter(lambda _: isinstance(_, (Args, Dict)), obj.values()):
        if (y := search_api(key, nested)) is not None:
            return y
    return None

def to_dict(args):
    rtn_dict = {}
    for k, v in args.items():
        if isinstance(v, Path):
            rtn_dict[k] = str(v)
        elif isinstance(v, Args):
            rtn_dict[k] = to_dict(v)
        rtn_dict[k] = v
    return rtn_dict


class Args(Mapping):

    def __init__(self, level=1, **kwargs):
        for key, value in kwargs.items():
            self.set(key, self.init_helper(value, level=level))
        if kwargs.pop('_post_init_', False):
            self.post_init()

    def to_string(self):
        return '_'.join([f'{k}={v}' for k,v in sorted(self.items())])
    
    @classmethod
    def init_helper(cls, value, level=1):
        if level > 1 and isinstance(value, (Dict, cls)):
            value = cls(level=level-1, **value)
        return value

    def search(self, key): #recursively find key in obj
        return search_api(key, self)

    def set(self, __name, __value):
        key, *subkey = __name.split('.', 1)
        if not subkey:
            setattr(self, key, __value)
        else:
            rtn_value = self.pop(key, Args())
            rtn_value.set(subkey[0], __value)
            setattr(self, key, rtn_value)

    def update(self, override=True, **kwargs):
        for k,v in kwargs.items(): 
            if not override and k in self:
                continue
            self.set(k, v)
        
    def pop(self,__name, __default=None) -> Any:
        return self.__dict__.pop(__name, __default)
    
    def pops(self, *__names, input_mappings={}):
        return {input_mappings.get(n, n): self.pop(n) for n in __names}

    def copy(self):
        return deepcopy(self)

    def __add__(self, other):
        return Args(**self, **other)

    def __contains__(self, item): return hasattr(self, item)
    
    def __iter__(self): yield from self.__dict__

    def get(self, item, *default): 
        if default: return getattr(self, item, default[0])
        else: return getattr(self, item)
    
    def __getitem__(self, item): return getattr(self, item)
    
    def __setitem__(self, item, value): setattr(self, item, value)

    def __len__(self): return len(self.__dict__)

    def post_init(self):
        parser = argparse.ArgumentParser()
        def f(args, prefix=''):
            flag = False
            for k,v in args.items():
                prefix_k = prefix + '.' + k
                if isinstance(v, (Args, dict)): f(v, prefix_k)
                elif v is None or v == 'None': args[k] = None
                elif isinstance(v, str) and v.startswith('$'):
                    if v.startswith('$$'): args[k] = v[1:]
                    else: args[k] = attrgetter(v.lstrip('$'))(self)
                elif isinstance(v, str) and v.startswith('--'):
                    if (match := re.match('--(\w+):\s*(.*)', v)):
                        _type, _value = match.groups()
                        preargs = [f"--{prefix_k[1:]}", f"-{k}"]
                        if _type == 'bool':
                            default = eval(_type)(_value)
                            parser.add_argument(*preargs, action='store_true' if default else 'store_false', default=not default)
                        elif _type in ['str', 'int', 'float']:
                            _type = eval(_type)
                            parser.add_argument(*preargs, type=_type, default=_type(_value))
                        elif _type.startswith('list'):
                            parser.add_argument(*preargs, nargs='*', default=eval(_value), type=eval(_type[5:]))
                        flag = True
                    else:
                        raise AttributeError("please make sure your argument follows --{type}:{default_value}")
            return flag
        if f(self):
            self.update(**vars(parser.parse_args()))

    def __repr__(self):
        return pformat(self.to_dict())

    def to_dict(self):
        return to_dict(self.copy())

    def __str__(self):
        choice = r"^(choice|interval|range|sort|shuffle|tag|glob)\(.*\)"
        def show(x):
            if x is None or x == 'null':
                return 'null'
            elif isinstance(x, str) and not re.match(choice, x): 
                return f'"{x}"'
            elif not isinstance(x, Mapping):
                return x
            return '{' + ', '.join(f"{k}: {show(v)}" for (k, v) in x.items()) + '}'
        return show(self)

    def hydra_prepare(self, sweep=False):
        """treat arg as sys.argv to be read by hydra"""
        if sweep: 
            sys.argv.extend(["--multirun", "hydra/sweeper=example"])
        for k,v in flatten_dict(self).items():  # type: ignore
            sys.argv.append("%s=%s"%(k, v))
    
    def parse_args(self, sep='', eql='=', ignore=[]):
        if sep != '':
            return sep.join([f'{k}{eql}{v}' for k,v in flatten_dict(self).items() if not k in ignore])
        from hydra.core.override_parser.overrides_parser import ( OverridesParser, create_functions,)
        rules = create_functions()
        parser = OverridesParser(rules)
        return parser.parse_rule(str(self), 'dictContainer')